/*
Copyright 2018 Harvey OS Team

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package main

import (
	"encoding/json"
	"flag"
	"io/ioutil"
	"log"
	"os"
	"path/filepath"
)

var config = struct {
	Harvey   string
	Args     []string
	Except   map[string]bool
	CmdName  string
	FullPath string
	Src      string
	Uroot    string
	Cwd      string
	Bbsh     string

	Goroot    string
	Gosrcroot string
	Arch      string
	Goos      string
	Gopath    string
	TempDir   string
	Go        string
	Debug     bool
	Fail      bool
}{
	Harvey: os.Getenv("HARVEY"),
	Except: map[string]bool{"cflags.json": true, "9": true, "libauthcmd.json": true, "awk.json": true, "bzip2": true, "klibc.json": true, "kcmds.json": true, "klib.json": true, "kernel.json": true},
}

type fixup func(string, map[string]interface{})

func cflags(n string, jsmap map[string]interface{}) {
	if _, ok := jsmap["Cflags"]; ok {
		log.Printf("Deleting Cflags from %v", n)
		delete(jsmap, "Cflags")
		// TODO: once we have another ARCH, use it.
		a := []string{"/amd64/include/cflags.json"}
		switch tval := jsmap["Include"].(type) {
		case []interface{}:
			for _, v := range tval {
				a = append(a, v.(string))
			}
		}
		jsmap["Include"] = a
	}
}

func removeempty(n string, jsmap map[string]interface{}) {
	for key, val := range jsmap {
		switch tval := val.(type) {
		case map[string]interface{}:
			log.Printf("%s: tval %s", n, tval)
			if len(tval) == 0 {
				delete(jsmap, key)
			}
		case []interface{}:
			if len(tval) == 0 {
				delete(jsmap, key)
			}
		}
	}
}

func checkname(n string, jsmap map[string]interface{}) {
	if _, ok := jsmap["Name"]; !ok {
		log.Printf("File %v has no \"Name\" key", n)
	}
}

func one(n string, f ...fixup) error {
	buf, err := ioutil.ReadFile(n)
	if err != nil {
		return err
	}

	var jsmap map[string]interface{}
	if err := json.Unmarshal(buf, &jsmap); err != nil {
		return err
	}
	if config.Debug {
		log.Printf("%v: %v", n, jsmap)
	}

	for _, d := range f {
		d(n, jsmap)
	}
	buf, err = json.MarshalIndent(jsmap, "", "\t")
	if err != nil {
		return err
	}
	buf = append(buf, '\n')

	if config.Debug {
		os.Stdout.Write(buf)
	} else {
		ioutil.WriteFile(n, buf, 0666)
	}
	return nil
}

func init() {
	if config.Harvey == "" {
		log.Fatalf("Please set $HARVEY")
	}

	if len(config.Args) == 0 {
		config.Args = []string{config.Harvey}
	}
}

func main() {
	var err error

	flag.BoolVar(&config.Debug, "d", true, "Enable debug prints")
	flag.Parse()
	config.Args = flag.Args()
	for _, n := range config.Args {
		err = filepath.Walk(n, func(name string, fi os.FileInfo, err error) error {
			if err != nil {
				return err
			}
			if fi.IsDir() {
				if config.Except[fi.Name()] {
					return filepath.SkipDir
				}
				return nil
			}
			n := fi.Name()
			if len(n) < 5 || n[len(n)-5:] != ".json" {
				return nil
			}
			todo := []fixup{cflags, removeempty, checkname}
			if config.Except[n] {
				todo = todo[1:]
			}
			log.Printf("process %s", name)
			err = one(name, todo...)
			if err != nil {
				log.Printf("%s: %s\n", name, err)
				return err
			}
			return nil
		})

		if err != nil {
			log.Fatalf("%v", err)

		}
	}
}
